# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import re

import venvs


# -- Project information -----------------------------------------------------

project = "venvs"
author = "Julian Berman"
copyright = "2015, " + author

release = venvs.__version__
version = release.partition("-")[0]


# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinxcontrib.spelling",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The reST default role (used for this markup: `text`) to use for all documents.
default_role = "any"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"https://docs.python.org/3/": None}

# -- Options for the linkcheck builder ------------------------------------


def entire_domain(host):
    return r"http.?://" + re.escape(host) + r"($|/.*)"


linkcheck_ignore = [
    entire_domain("codecov.io"),
    "https://github.com/Julian/venvs/actions",
    "https://github.com/Julian/venvs/workflows/CI/badge.svg",
]

linkcheck_anchors = False

# -- Options for sphinxcontrib-spelling -----------------------------------

spelling_word_list_filename = "spelling-wordlist.txt"



from datetime import datetime, timedelta
from tempfile import NamedTemporaryFile
from unittest import TestCase
import os

from filesystems.exceptions import FileExists

from venvs import _config
from venvs.tests.utils import CLIMixin


class TestConverge(CLIMixin, TestCase):
    def test_it_creates_missing_virtualenvs(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))
        self.assertFalse(self.locator.for_name("b").exists_on(self.filesystem))
        self.assertFalse(self.locator.for_name("c").exists_on(self.filesystem))

        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            [virtualenv.b]
            install = ["foo", "bar", "bla"]
            requirements = ["requirements.txt"]
            [virtualenv.c]
            install = ["foo", "$HOME", "~/a"]
            link = ["bar", "baz"]
            """
        )

        self.run_cli(["converge"])

        self.assertEqual(
            (
                self.installed(self.locator.for_name("a")),
                self.installed(self.locator.for_name("b")),
                self.installed(self.locator.for_name("c")),
            ), (
                (set(), set()),
                ({"foo", "bar", "bla"}, {"requirements.txt"}),
                (
                    {
                        "foo",
                        os.path.expandvars("$HOME"),
                        os.path.expanduser("~/a"),
                    },
                    set(),
                ),
            ),
        )

    def test_it_converges_existing_virtualenvs(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))
        self.assertFalse(self.locator.for_name("b").exists_on(self.filesystem))
        self.assertFalse(self.locator.for_name("c").exists_on(self.filesystem))

        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            install = ["foo", "bar"]
            requirements = ["requirements.txt"]
            """
        )

        self.run_cli(["converge"])

        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            install = ["baz", "quux"]
            requirements = ["requirements.txt", "other.txt"]
            """
        )

        self.run_cli(["converge"])

        self.assertEqual(
            self.installed(self.locator.for_name("a")),
            ({"baz", "quux"}, {"requirements.txt", "other.txt"}),
        )

    def test_it_converges_specified_virtualenvs(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))
        self.assertFalse(self.locator.for_name("b").exists_on(self.filesystem))
        self.assertFalse(self.locator.for_name("c").exists_on(self.filesystem))

        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            [virtualenv.b]
            install = ["foo", "bar", "bla"]
            requirements = ["requirements.txt"]
            [virtualenv.c]
            install = ["foo"]
            link = ["bar", "baz"]
            """
        )

        self.run_cli(["converge", "a", "c"])

        self.assertTrue(self.locator.for_name("a").exists_on(self.filesystem))
        self.assertFalse(self.locator.for_name("b").exists_on(self.filesystem))
        self.assertTrue(self.locator.for_name("c").exists_on(self.filesystem))

        self.assertEqual(
            (
                self.installed(self.locator.for_name("a")),
                self.installed(self.locator.for_name("b")),
                self.installed(self.locator.for_name("c")),
            ), (
                (set(), set()),
                (set(), set()),
                ({"foo"}, set()),
            ),
        )

    def test_it_runs_post_commands(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))

        file = NamedTemporaryFile(delete=False)
        self.addCleanup(os.remove, file.name)
        mtime = os.path.getmtime(file.name)
        new_mtime = datetime.fromtimestamp(mtime) + timedelta(minutes=10)

        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), f"""
            [virtualenv.a]
            post-commands = [
                ["true"],
                ["touch", "-t", "{new_mtime:%Y%m%d%H%M}", "{file.name}"],
            ]
            """
        )

        self.run_cli(["converge"])

        self.assertTrue(self.locator.for_name("a").exists_on(self.filesystem))
        self.assertGreater(os.path.getmtime(file.name), mtime)

    def test_it_does_not_run_post_commands_for_already_converged_envs(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))

        file = NamedTemporaryFile(delete=False)
        self.addCleanup(os.remove, file.name)

        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), f"""
            [virtualenv.a]
            post-commands = [
                ["true"],
                ["touch", "{file.name}"],
            ]
            """
        )

        self.run_cli(["converge"])
        mtime = os.path.getmtime(file.name)

        self.run_cli(["converge"])
        self.assertEqual(os.path.getmtime(file.name), mtime)

    def test_it_stops_post_commands_on_error(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))

        file = NamedTemporaryFile(delete=False)
        self.addCleanup(os.remove, file.name)
        mtime = os.path.getmtime(file.name)

        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), f"""
            [virtualenv.a]
            post-commands = [
                ["false"],
                ["touch", "{file.name}"],
            ]
            """
        )

        self.run_cli(["converge"])

        self.assertEqual(os.path.getmtime(file.name), mtime)

    def test_bundles(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))

        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [bundle]
            dev = ["bar", "bla"]

            [virtualenv.a]
            install = ["foo"]
            install-bundle = ["dev"]
            """
        )

        self.run_cli(["converge"])

        self.assertEqual(
            self.installed(self.locator.for_name("a")),
            ({"foo", "bar", "bla"}, set()),
        )

    def test_modifying_a_bundle_recreates_envs_using_it(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [bundle]
            dev = ["bar"]

            [virtualenv.a]
            install-bundle = ["dev"]
            """
        )

        self.run_cli(["converge"])
        self.assertEqual(
            self.installed(self.locator.for_name("a")),
            ({"bar"}, set()),
        )

        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [bundle]
            dev = ["bar", "baz"]

            [virtualenv.a]
            install-bundle = ["dev"]
            """
        )

        self.run_cli(["converge"])

        self.assertEqual(
            self.installed(self.locator.for_name("a")),
            ({"bar", "baz"}, set()),
        )

    def test_no_such_bundle(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))

        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            install = ["foo"]
            install-bundle = ["dev"]
            """
        )

        self.run_cli(["converge"])

        self.assertEqual(
            self.installed(self.locator.for_name("a")),
            (set(), set()),
        )

    def test_it_does_not_blow_up_by_default_on_install(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            [virtualenv.b]
            [virtualenv.magicExplodingVirtualenvOnInstall]
            [virtualenv.c]
            """
        )

        self.run_cli(["converge"])

        self.assertEqual(
            (
                self.installed(self.locator.for_name("a")),
                self.installed(self.locator.for_name("b")),
                self.installed(self.locator.for_name("c")),
                self.locator.for_name("c").exists_on(self.filesystem),
            ),
            tuple((set(), set()) for _ in "abc") + (True,),
        )
        self.assertIn(
            "'magicExplodingVirtualenvOnInstall' failed",
            self.stderr.getvalue(),
        )

    def test_it_can_be_asked_to_blow_up_immediately_on_install(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            [virtualenv.b]
            [virtualenv.magicExplodingVirtualenvOnInstall]
            [virtualenv.c]
            """
        )

        with self.assertRaises(ZeroDivisionError):
            self.run_cli(["converge", "--fail-fast"])

        self.assertEqual(
            (
                self.installed(self.locator.for_name("a")),
                self.installed(self.locator.for_name("b")),
                self.locator.for_name("c").exists_on(self.filesystem),
            ),
            ((set(), set()), (set(), set()), False),
        )

    def test_it_does_not_blow_up_by_default_on_create(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            [virtualenv.b]
            [virtualenv.magicExplodingVirtualenvOnCreate]
            [virtualenv.c]
            """
        )

        self.run_cli(["converge"])

        self.assertEqual(
            (
                self.installed(self.locator.for_name("a")),
                self.installed(self.locator.for_name("b")),
                self.installed(self.locator.for_name("c")),
                self.locator.for_name("c").exists_on(self.filesystem),
            ),
            tuple((set(), set()) for _ in "abc") + (True,),
        )
        self.assertIn(
            "'magicExplodingVirtualenvOnCreate' failed",
            self.stderr.getvalue(),
        )

    def test_it_can_be_asked_to_blow_up_immediately_on_create(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            [virtualenv.b]
            [virtualenv.magicExplodingVirtualenvOnCreate]
            [virtualenv.c]
            """
        )

        with self.assertRaises(ZeroDivisionError):
            self.run_cli(["converge", "--fail-fast"])

        self.assertEqual(
            (
                self.installed(self.locator.for_name("a")),
                self.installed(self.locator.for_name("b")),
                self.locator.for_name("c").exists_on(self.filesystem),
            ),
            ((set(), set()), (set(), set()), False),
        )

    def test_specified_python(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            python = "python3"
            """
        )

        self.run_cli(["converge"])

        # FIXME: this doesn't properly assert about the python version...
        self.assertEqual(
            self.installed(self.locator.for_name("a")),
            (set(), set()),
        )

    def test_custom_link_dir(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link = ["foo"]
            """
        )

        link_dir = self.link_dir.descendant("some", "child", "bin")
        self.run_cli(["converge", "--link-dir", str(link_dir)])

        self.assertEqual(
            self.filesystem.readlink(link_dir / "foo"),
            self.locator.for_name("a").binary("foo"),
        )

    def test_link_exists(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link = ["foo"]
            """
        )

        self.filesystem.touch(self.link_dir.descendant("foo"))

        with self.assertRaises(FileExists):
            self.run_cli(["converge"])

    def test_link_exists_as_broken_symlink(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link = ["foo"]
            """
        )

        self.filesystem.link(
            source=self.link_dir.descendant("broken"),
            to=self.link_dir.descendant("foo"),
        )

        self.run_cli(["converge"])

        self.assertEqual(
            self.linked,
            {"foo": self.locator.for_name("a").binary("foo")},
        )

    def test_conflicting_links(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link = ["foo"]

            [virtualenv.b]

            [virtualenv.c]
            link = ["foo"]
            """
        )

        with self.assertRaises(_config.DuplicatedLinks) as e:
            self.run_cli(["converge"])

        self.assertIn("foo", str(e.exception))
        self.assertEqual(self.linked, {})

    def test_conflicting_links_via_rename(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link = ["foo"]

            [virtualenv.b]

            [virtualenv.c]
            link = ["bar:foo"]
            """
        )

        with self.assertRaises(_config.DuplicatedLinks) as e:
            self.run_cli(["converge"])

        self.assertIn("foo", str(e.exception))
        self.assertEqual(self.linked, {})

    def test_specified_link_name(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link = ["foo:fooBar"]
            """
        )

        self.run_cli(["converge"])

        self.assertEqual(
            self.linked,
            {"fooBar": self.locator.for_name("a").binary("foo")},
        )

    def test_missing_link_dir(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link = ["foo"]
            """
        )

        self.filesystem.remove_empty_directory(self.link_dir)
        self.assertFalse(self.filesystem.is_dir(path=self.link_dir))

        self.run_cli(["converge"])

        self.assertEqual(
            self.linked,
            {"foo": self.locator.for_name("a").binary("foo")},
        )

    def test_link_m_module(self):
        """
        It links modules run via -m as wrappers.
        """
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link-module = ["this"]
            """
        )

        self.run_cli(["converge"])

        contents = self.filesystem.get_contents(
            self.link_dir.descendant("this"),
        )
        self.assertEqual(
            contents.splitlines()[0],
            "#!" + str(self.locator.for_name("a").binary("python")),
        )

    def test_link_m_module_specified_name(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link-module = ["this:that"]
            """
        )

        self.run_cli(["converge"])

        contents = self.filesystem.get_contents(
            self.link_dir.descendant("that"),
        )
        self.assertEqual(
            contents.splitlines()[0],
            "#!" + str(self.locator.for_name("a").binary("python")),
        )

    def test_link_m_module_duplicated(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link = ["foo"]

            [virtualenv.b]

            [virtualenv.c]
            link-module = ["bar:foo"]
            """
        )

        with self.assertRaises(_config.DuplicatedLinks) as e:
            self.run_cli(["converge"])

        self.assertIn("foo", str(e.exception))
        self.assertEqual(self.linked, {})

    def test_link_m_module_replaces_generated_files(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link-module = ["this"]
            """
        )

        self.run_cli(["converge"])

        # Just change the config in a way that will re-converge
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link-module = ["this", "that"]
            """
        )

        self.run_cli(["converge"])

    def test_link_m_module_does_not_replace_non_venvs_wrappers(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            link-module = ["this"]
            """
        )

        self.filesystem.touch(self.link_dir.descendant("this"))

        with self.assertRaises(FileExists):
            self.run_cli(["converge"])

    def test_changing_a_bundle_recreates_the_venv(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [bundle]
            one = ["foo"]

            [virtualenv.a]
            install-bundle = ["one"]
            """
        )
        self.run_cli(["converge"])
        self.assertEqual(
            self.installed(self.locator.for_name("a")),
            ({"foo"}, set()),
        )

        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [bundle]
            one = ["foo", "bar"]

            [virtualenv.a]
            install-bundle = ["one"]
            """
        )

        self.run_cli(["converge"])

        self.assertEqual(
            self.installed(self.locator.for_name("a")),
            ({"foo", "bar"}, set()),
        )

    def test_missing_config_recreates_the_venv(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            """
        )
        self.run_cli(["converge"])

        venv = self.locator.for_name("a")
        self.filesystem.remove_file(venv.path / "installed.json")

        some_random_file = venv.path / "some-random-file"
        self.filesystem.touch(some_random_file)
        self.assertTrue(self.filesystem.is_file(some_random_file))

        # Now the file should disappear as the venv gets recreated
        self.run_cli(["converge"])
        self.assertFalse(self.filesystem.is_file(some_random_file))

    def test_invalid_config_recreates_the_venv(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            """
        )
        self.run_cli(["converge"])

        venv = self.locator.for_name("a")
        self.filesystem.set_contents(
            venv.path / "installed.json",
            "not even json",
        )

        some_random_file = venv.path / "some-random-file"
        self.filesystem.touch(some_random_file)
        self.assertTrue(self.filesystem.is_file(some_random_file))

        # Now the file should disappear as the venv gets recreated
        self.run_cli(["converge"])
        self.assertFalse(self.filesystem.is_file(some_random_file))

    def test_valid_json_invalid_config_recreates_the_venv(self):
        self.filesystem.set_contents(
            self.locator.root.descendant("virtualenvs.toml"), """
            [virtualenv.a]
            """
        )
        self.run_cli(["converge"])

        venv = self.locator.for_name("a")
        self.filesystem.set_contents(venv.path / "installed.json", "{}")

        some_random_file = venv.path / "some-random-file"
        self.filesystem.touch(some_random_file)
        self.assertTrue(self.filesystem.is_file(some_random_file))

        # Now the file should disappear as the venv gets recreated
        self.run_cli(["converge"])
        self.assertFalse(self.filesystem.is_file(some_random_file))



"""
Converge the set of installed virtualenvs.

"""
from datetime import datetime
import subprocess
import sys

from filesystems.exceptions import FileExists, FileNotFound
from tqdm import tqdm
import click

from venvs import __version__
from venvs._config import Config
from venvs.common import _FILESYSTEM, _LINK_DIR, _ROOT


_SENTINEL = "# Auto-generated by venvs"
_MODULE_WRAPPER = """\
#!{python}
""" + _SENTINEL + """ v{version} on {now}.
# Don't modify this file, it may be replaced when re-converging.
import os
import sys
# No runpy.run_module because of https://bugs.python.org/issue37941 ...
argv = [sys.executable, "-m", "{module}"] + sys.argv[1:]
os.execvp(argv[0], argv)
"""


def _fail(virtualenv, name):
    raise


def _do_not_fail(virtualenv, name):
    sys.stderr.write("Converging {!r} failed!\n".format(name))


@_FILESYSTEM
@_LINK_DIR
@_ROOT
@click.option(
    "--fail-fast", "handle_error",
    flag_value=_fail,
    help="Fail if any virtualenv cannot be converged.",
)
@click.option(
    "--no-fail-fast", "handle_error",
    default=True,
    flag_value=_do_not_fail,
    help="Do not fail if a virtualenv cannot be converged.",
)
@click.argument("venvs", nargs=-1)
def main(filesystem, locator, link_dir, handle_error, venvs):
    """
    Converge the configured set of tracked virtualenvs.
    """
    for config, virtualenv in _loop(
        filesystem=filesystem,
        locator=locator,
        handle_error=handle_error,
    ):
        if venvs and config.name not in venvs:
            continue

        try:
            virtualenv.recreate_on(filesystem=filesystem, python=config.python)
            virtualenv.install(
                packages=config.install,
                requirements=config.requirements,
            )
        except Exception:
            handle_error(virtualenv=virtualenv, name=config.name)
            continue

        filesystem.create_directory(
            link_dir,
            with_parents=True,
            allow_existing=True,
        )

        for name, to in config.link.items():
            _link(
                source=virtualenv.binary(name=name),
                to=link_dir.descendant(to),
                filesystem=filesystem,
            )

        for name, to in config.link_module.items():
            _write_module_wrapper(
                to=link_dir.descendant(to),
                python=virtualenv.binary(name="python"),
                module=name,
                filesystem=filesystem,
            )

        config.save(filesystem=filesystem, virtualenv=virtualenv)

        try:
            for command in config.post_commands:
                subprocess.run(command, check=True)
        except Exception:
            handle_error(virtualenv=virtualenv, name=config.name)
            continue


def _loop(filesystem, locator, handle_error):
    config = Config.from_locator(filesystem=filesystem, locator=locator)
    progress = tqdm(iterable=config, total=len(config), unit="venv")
    iterable = iter(progress)
    while True:
        try:
            venv_config = next(iterable)
        except StopIteration:
            return
        except Exception:
            handle_error(virtualenv=None, name=None)
        else:
            progress.set_description(venv_config.name)
            venv = locator.for_name(name=venv_config.name)
            if venv_config.matches_existing(venv, filesystem=filesystem):
                continue
            yield venv_config, venv


def _link(source, to, filesystem):
    """
    Link the given binary, replacing broken symlinks and erroring if existing.
    """

    try:
        filesystem.link(source=source, to=to)
    except FileExists as error:
        if filesystem.realpath(error.value) == filesystem.realpath(source):
            return
        if filesystem.exists(to):
            raise
        filesystem.remove(to)
        filesystem.link(source=source, to=to)


def _write_module_wrapper(to, python, module, filesystem):
    """
    Write a module wrapper for the given module.

    A module wrapper converts `python -m foo` into a bare `foo` executable.
    """
    wrapper = _MODULE_WRAPPER.format(
        python=python,
        module=module,
        version=__version__,
        now=datetime.now(),
    )

    try:
        with filesystem.open(to) as file:
            file.readline()  # shebang
            second_line = file.readline()
    except FileNotFound:
        pass
    else:
        if not second_line.startswith(_SENTINEL):
            raise FileExists(to)

    filesystem.set_contents(to, wrapper)



import os
from unittest import TestCase

from filesystems import Path

from venvs.tests.utils import CLIMixin


class TestFind(CLIMixin, TestCase):
    def test_find_directory_finds_envs_by_directory(self):
        this_dir = Path.from_string(__file__).parent()
        stdout, stderr = self.run_cli(["find", "directory", str(this_dir)])
        self.assertEqual(
            (stdout, stderr),
            (str(self.locator.for_directory(this_dir).path) + "\n", ""),
        )

    def test_find_directory_defaults_to_cwd(self):
        stdout, stderr = self.run_cli(["find", "directory"])
        self.assertEqual(
            (stdout, stderr), (
                str(self.locator.for_directory(Path.cwd()).path) + "\n",
                "",
            ),
        )

    def test_find_name_finds_envs_by_name(self):
        stdout, stderr = self.run_cli(["find", "name", "bla"])
        self.assertEqual(
            (stdout, stderr),
            (str(self.locator.for_name("bla").path) + "\n", ""),
        )

    def test_find_without_args_finds_the_virtualenv_root(self):
        stdout, stderr = self.run_cli(["find"])
        self.assertEqual(
            (stdout, stderr),
            (str(self.locator.root) + os.linesep, ""),
        )

    def test_find_directory_with_binary(self):
        this_dir = Path.from_string(__file__).parent()
        stdout, stderr = self.run_cli(
            ["find", "directory", str(this_dir), "python"],
        )
        this_dir_venv = self.locator.for_directory(this_dir)
        self.assertEqual(
            (stdout, stderr),
            (str(this_dir_venv.binary("python")) + "\n", ""),
        )

    def test_find_existing_by_name_fails_for_non_existing_virtualenvs(self):
        stdout, stderr = self.run_cli(
            ["find", "--existing-only", "name", "bla"], exit_status=1,
        )
        self.assertEqual((stdout, stderr), ("", ""))

    def test_find_existing_by_name_succeeds_for_existing_virtualenvs(self):
        self.filesystem.create_directory(self.locator.root / "bla")

        stdout, stderr = self.run_cli(
            ["find", "--existing-only", "name", "bla"],
        )
        self.assertEqual(
            (stdout, stderr),
            (str(self.locator.for_name("bla").path) + "\n", ""),
        )

    def test_find_existing_by_dir_fails_for_non_existing_virtualenvs(self):
        stdout, stderr = self.run_cli(
            ["find", "--existing-only", "directory", "/foo/bla"],
            exit_status=1,
        )
        self.assertEqual((stdout, stderr), ("", ""))

    def test_find_existing_by_dir_succeeds_for_existing_virtualenvs(self):
        self.filesystem.create_directory(self.locator.root / "bla")

        path = Path("foo", "bla")
        stdout, stderr = self.run_cli(
            ["find", "--existing-only", "directory", str(path)],
        )
        self.assertEqual(
            (stdout, stderr),
            (str(self.locator.for_directory(path).path) + "\n", ""),
        )

    def test_cannot_specify_random_garbage(self):
        stdout, stderr = self.run_cli(
            ["find", "--random-garbage"], exit_status=2,
        )
        stderr_ends_with = "Error: no such option: --random-garbage\n"
        self.assertEqual(
            (stdout, stderr[-len(stderr_ends_with):]),
            (stdout, stderr_ends_with),
        )



from unittest import TestCase

from venvs.tests.utils import CLIMixin


class TestTemporary(CLIMixin, TestCase):
    def test_it_creates_a_global_temporary_environment(self):
        temporary = self.locator.temporary()
        self.assertFalse(temporary.exists_on(self.filesystem))

        stdout, stderr = self.run_cli(["temporary"])
        self.assertEqual(
            (temporary.exists_on(self.filesystem), stdout, stderr),
            (True, str(temporary.path / "bin") + "\n", ""),
        )

    def test_it_recreates_the_environment_if_it_exists(self):
        temporary = self.locator.temporary()
        self.assertFalse(temporary.exists_on(self.filesystem))
        self.run_cli(["temporary"])
        self.assertTrue(temporary.exists_on(self.filesystem))

        foo = temporary.path / "foo"
        self.filesystem.touch(path=foo)
        self.assertTrue(self.filesystem.exists(path=foo))

        self.run_cli(["temporary"])
        self.assertTrue(temporary.exists_on(self.filesystem))
        self.assertFalse(self.filesystem.exists(path=foo))

    def test_env_with_single_install(self):
        self.run_cli(["temporary", "-i", "thing"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.temporary()), ({"thing"}, set()),
        )



from filesystems.exceptions import FileNotFound
import click

from venvs.common import _FILESYSTEM, _ROOT, _EX_NOINPUT


def run(locator, filesystem, names, force):
    for name in names:
        virtualenv = locator.for_name(name=name)
        try:
            virtualenv.remove_from(filesystem=filesystem)
        except FileNotFound:
            if not force:
                return _EX_NOINPUT


@_FILESYSTEM
@_ROOT
@click.option(
    "-f", "--force",
    flag_value=True,
    help="Ignore errors if the virtualenv does not exist.",
)
@click.argument("names", nargs=-1)
@click.pass_context
def main(context, **kwargs):
    """
    Remove an ad hoc virtualenv.
    """
    context.exit(run(**kwargs) or 0)



import click

from venvs import __version__, converge, create, find, remove, temporary


@click.group(context_settings=dict(help_option_names=["--help", "-h"]))
@click.version_option(prog_name="venvs", version=__version__)
def main():
    """
    Centralized virtual environments.
    """


main.command(name="converge")(converge.main)
main.command(name="create")(create.main)
main.add_command(find.main, name="find")
main.command(name="remove")(remove.main)
main.command(name="temporary")(temporary.main)



import json
import os
import subprocess
import sys

try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache

from pyrsistent import freeze, pmap, pvector, thaw
import attr
import filesystems.exceptions
import tomlkit


class DuplicatedLinks(Exception):
    pass


def _empty():
    table = tomlkit.table()
    table["bundle"] = tomlkit.table()
    table["virtualenv"] = tomlkit.table()
    return table


@attr.s(frozen=True)
class ConfiguredVirtualEnv(object):
    """
    A virtual environment defined within a config file section.
    """

    name = attr.ib()
    python = attr.ib(default=sys.executable)
    install = attr.ib(default=pvector())
    requirements = attr.ib(default=pvector())
    link = attr.ib(default=pmap())
    link_module = attr.ib(default=pmap())
    post_commands = attr.ib(default=pvector())

    @classmethod
    def from_dict(cls, name, config_dict, bundles):
        # tomlkit's data structures are broken in at least one way,
        # see sdipater/tomlkit#49, but I don't trust them not to be
        # broken in other ways given that they inherit from dict
        requirements = _interpolated(config_dict.get("requirements", []))
        install = list(_interpolated(config_dict.get("install", [])))

        for bundle_name in config_dict.get("install-bundle", []):
            bundle = bundles[bundle_name]
            for each in bundle:
                if each not in install:
                    install.append(each)

        kwargs = dict(
            install=pvector(install),
            requirements=pvector(requirements),
            python=config_dict.get("python", sys.executable),
            post_commands=freeze(config_dict.get("post-commands", [])),
        )
        for section in "link", "link-module":
            links = (
                each.partition(":") for each in config_dict.get(section, [])
            )
            kwargs[section.replace("-", "_")] = pmap(
                (name, to or name) for name, _, to in links
            )
        return cls(name=name, **kwargs)

    def save(self, filesystem, virtualenv):
        filesystem.set_contents(
            virtualenv.path / "installed.json",
            mode="t",
            contents=json.dumps(
                self._serializable(),
                ensure_ascii=False,
                indent=2,
            ),
        )

    def matches_existing(self, virtualenv, filesystem):
        try:
            existing = json.loads(
                filesystem.get_contents(virtualenv.path / "installed.json")
            )
        except (ValueError, filesystems.exceptions.FileNotFound):
            return False
        return existing == self._serializable()

    def _serializable(self):
        return {
            "virtualenv": thaw(pmap(attr.asdict(self))),
            "sys.version": _version_of(self.python),
        }


@attr.s(eq=False, frozen=True)
class Config(object):
    """
    A converge configuration file.
    """

    _contents = attr.ib(factory=_empty)

    @classmethod
    def from_string(cls, string):
        document = tomlkit.loads(string)
        # tomlkit's setdefault is broken, see sdipater/tomlkit#49
        for each in "virtualenv", "bundle":
            if each not in document:
                document[each] = tomlkit.table()
        _check_for_duplicated_links(document["virtualenv"].values())
        return cls(contents=document)

    @classmethod
    def from_locator(cls, locator, filesystem):
        contents = filesystem.get_contents(
            locator.root.descendant("virtualenvs.toml"),
            mode="t",
        )
        return cls.from_string(contents)

    def __iter__(self):
        return (
            ConfiguredVirtualEnv.from_dict(
                name=name,
                config_dict=config,
                bundles=self._contents["bundle"]
            ) for name, config in self._contents["virtualenv"].items()
        )

    def __eq__(self, other):
        """
        Compare the effective configuration dictated by the 2 configs.
        """
        if self.__class__ is not other.__class__:
            return NotImplemented
        return all(a == b for a, b in zip(self, other))

    def __ne__(self, other):
        return not self == other

    def __len__(self):
        return len(self._contents["virtualenv"])

    def add(
        self,
        name,
        install=(),
        requirements=(),
        link=(),
        link_module=(),
    ):
        # tomlkit's table.copy is broken, see sdipater/tomlkit#65
        contents = tomlkit.parse(self._contents.as_string())
        contents["virtualenv"].add(name, tomlkit.table())
        contents["virtualenv"][name].update(
            {
                "install": list(install),
                "requirements": list(requirements),
                "link": sorted(set(link)),
                "link-module": sorted(set(link_module)),
            },
        )
        return attr.evolve(self, contents=contents)


def _interpolated(iterable):
    return (os.path.expandvars(os.path.expanduser(each)) for each in iterable)


def add_virtualenv(filesystem, locator, installs, links, name):
    try:
        config = Config.from_locator(filesystem=filesystem, locator=locator)
    except filesystems.exceptions.FileNotFound:
        config = Config()
    new = config.add(name, install=installs, link=links)
    filesystem.set_contents(
        locator.root.descendant("virtualenvs.toml"),
        tomlkit.dumps(new._contents),
        mode="t",
    )


def _check_for_duplicated_links(sections):
    seen, duplicated = set(), set()
    for each in sections:
        for link in each.get("link", []) + each.get("link-module", []):
            name, _, to = link.partition(":")
            to = to or name
            if to in seen:
                duplicated.add(to)
            seen.add(to)
    if duplicated:
        raise DuplicatedLinks(duplicated)


@lru_cache()
def _version_of(python):
    return subprocess.check_output(
        [python, "--version"],
        stderr=subprocess.STDOUT,
    ).decode("ascii")



from unittest import TestCase
import sys

from filesystems import Path
import filesystems.native

from venvs import _cli, _config
from venvs.common import Locator
from venvs.tests.utils import CLIMixin


class TestCreate(CLIMixin, TestCase):
    def test_create_creates_an_env_with_the_given_name(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))
        self.run_cli(["create", "a"])
        self.assertTrue(self.locator.for_name("a").exists_on(self.filesystem))

    def test_recreate(self):
        virtualenv = self.locator.for_name("something")
        self.assertFalse(virtualenv.exists_on(self.filesystem))

        virtualenv.create()

        thing = virtualenv.path / "thing"
        self.filesystem.touch(path=thing)
        self.assertTrue(self.filesystem.exists(thing))

        self.run_cli(["create", "--recreate", "something"])
        self.assertTrue(virtualenv.exists_on(self.filesystem))

        self.assertFalse(self.filesystem.exists(thing))

    def test_install_and_requirements(self):
        self.run_cli(
            ["create", "-i", "foo", "-i", "bar", "-r", "reqs.txt", "bla"],
        )
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.for_name("bla")),
            ({"foo", "bar"}, {"reqs.txt"}),
        )

    def test_venvs_default_name(self):
        """
        Just saying ``venvs`` creates an environment based on the current
        directory's name.
        """

        virtualenv = self.locator.for_directory(Path.cwd())
        self.assertFalse(virtualenv.exists_on(self.filesystem))
        self.run_cli(["create"])
        self.assertTrue(virtualenv.exists_on(self.filesystem))

    def test_install_default_name(self):
        """
        If you install one single package and don't specify a name, the name of
        the installed package is used.
        """

        self.run_cli(["create", "-i", "foo"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.for_name("foo")), ({"foo"}, set()),
        )

    def test_install_default_name_with_version_specification(self):
        self.run_cli(["create", "-i", "thing[foo]>=2,<3"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.for_name("thing")),
            ({"thing[foo]>=2,<3"}, set()),
        )

    def test_link_default_name(self):
        """
        If you link one single binary and don't specify a name, the name of
        the binary is probably both the package and what you want to call the
        environment.

        """

        self.run_cli(["create", "-l", "foo"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.for_name("foo")), ({"foo"}, set()),
        )

    def test_multiple_installs_one_link(self):
        self.run_cli(["create", "-i", "foo", "-i", "bar", "-l", "foo", "baz"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.for_name("baz")),
            ({"foo", "bar"}, set()),
        )

    def test_multiple_installs_one_link_no_name(self):
        self.run_cli(
            ["create", "-i", "foo", "-i", "bar", "-l", "foo"],
            exit_status=2,
        )

    def test_install_edit_config(self):
        """Install --persist edits the config file."""
        self.run_cli(["create", "-l", "foo", "-i", "bar", "--persist"])

        self.assertConfigEqual(
            """
            [virtualenv.bar]
            install = ["bar"]
            link = ["foo"]
            """,
        )

    def test_handle_empty_config_file(self):
        """Don't break with an empty config file."""

        self.filesystem.touch(self.locator.root.descendant("virtualenvs.toml"))

        self.run_cli(["create", "-l", "foo", "-i", "bar", "--persist"])

        self.assertConfigEqual(
            """
            [virtualenv.bar]
            install = ["bar"]
            link = ["foo"]
            """,
        )

    def test_persist_handles_missing_config_directory(self):
        """Create the config directory if it does not exist."""

        self.filesystem.remove_empty_directory(self.locator.root)

        self.run_cli(["create", "-l", "foo", "-i", "bar", "--persist"])

        self.assertConfigEqual(
            """
            [virtualenv.bar]
            install = ["bar"]
            link = ["foo"]
            """,
        )

    def test_no_persist_handles_missing_virtualenv_directory(self):
        """Don't break if there is no virtualenv directory."""

        self.filesystem.remove_empty_directory(self.locator.root)

        self.run_cli(["create", "-l", "foo", "-i", "bar", "--no-persist"])

        self.assertTrue(self.filesystem.exists(self.locator.root))

        # Config file has _not_ been created.
        self.assertFalse(
            self.filesystem.exists(
                self.locator.root.descendant("virtualenvs.toml")
            )
        )

    def test_install_no_persist(self):
        """Install --no-persist does not edit the config file."""
        self.run_cli(["create", "-l", "foo", "-i", "bar", "--no-persist"])

        # No file has been created.
        with self.assertRaises(filesystems.exceptions.FileNotFound):
            _config.Config.from_locator(
                filesystem=self.filesystem,
                locator=self.locator,
            )


class TestIntegration(TestCase):
    def setUp(self):
        self.fs = filesystems.native.FS()
        self.root = self.fs.temporary_directory()
        self.addCleanup(self.fs.remove, self.root)

        # Fucking click.
        stdout = sys.stdout
        self.addCleanup(lambda: setattr(sys, "stdout", stdout))

    def test_it_works(self):
        with self.fs.open(self.root / "create_stdout", "w") as stdout:
            sys.stdout = stdout

            try:
                _cli.main(
                    args=[
                        "create",
                        "--root", str(self.root),
                        "venvs-unittest-should-be-deleted",
                    ],
                )
            except SystemExit:
                pass

        with self.fs.open(self.root / "find_stdout", "w") as stdout:
            sys.stdout = stdout

            try:
                _cli.main(
                    [
                        "find",
                        "--root", str(self.root),
                        "--existing-only",
                        "name", "venvs-unittest-should-be-deleted",
                    ],
                )
            except SystemExit:
                pass

        locator = Locator(root=self.root)
        virtualenv = locator.for_name("venvs-unittest-should-be-deleted")
        self.assertEqual(
            self.fs.get_contents(self.root / "find_stdout"),
            str(virtualenv.path) + "\n",
        )



import click

from venvs.common import _FILESYSTEM, _ROOT
from venvs.create import _INSTALL, _REQUIREMENTS


@_FILESYSTEM
@_ROOT
@_INSTALL
@_REQUIREMENTS
def main(filesystem, locator, installs, requirements):
    """
    Create or reuse the global temporary virtualenv.
    """

    temporary = locator.temporary()
    click.echo(temporary.binary("python").dirname())
    temporary.recreate_on(filesystem=filesystem)
    temporary.install(packages=installs, requirements=requirements)



from unittest import TestCase
import os

import tomlkit.exceptions

from venvs import _config


class TestConfig(TestCase):
    def test_simple(self):
        config = _config.Config.from_string(
            """
            [virtualenv.a]
            [virtualenv.b]
            install = ["foo", "bar", "bla"]
            """,
        )
        self.assertEqual(
            list(config), [
                _config.ConfiguredVirtualEnv(name="a"),
                _config.ConfiguredVirtualEnv(
                    name="b",
                    install=["foo", "bar", "bla"],
                ),
            ],
        )

    def test_links(self):
        config = _config.Config.from_string(
            """
            [virtualenv.a]
            link = ["foo", "bar", "baz:quux"]
            link-module = ["spam", "eggs:cheese"]
            """,
        )
        self.assertEqual(
            list(config), [
                _config.ConfiguredVirtualEnv(
                    name="a",
                    link={"foo": "foo", "bar": "bar", "baz": "quux"},
                    link_module={"spam": "spam", "eggs": "cheese"},
                ),
            ],
        )

    def test_bundles(self):
        config = _config.Config.from_string(
            """
            [bundle]
            one = ["foo", "bar"]

            [virtualenv.a]
            install-bundle = ["one"]

            [virtualenv.b]
            install-bundle = ["one"]
            install = ["bar", "baz"]
            """,
        )
        self.assertEqual(
            list(config), [
                _config.ConfiguredVirtualEnv(name="a", install=["foo", "bar"]),
                _config.ConfiguredVirtualEnv(
                    name="b",
                    install=["bar", "baz", "foo"],
                ),
            ],
        )

    def test_no_such_bundle(self):
        config = _config.Config.from_string(
            """
            [virtualenv.a]
            install-bundle = ["one"]
            """,
        )
        with self.assertRaises(tomlkit.exceptions.NonExistentKey) as e:
            next(iter(config))
        self.assertIn('"one"', str(e.exception))

    def test_expansion(self):
        config = _config.Config.from_string(
            """
            [virtualenv.a]
            install = ["~/a", "$HOME", "${HOME}/b"]
            requirements = ["requirements-${HOME}.txt"]
            """,
        )
        self.assertEqual(
            list(config), [
                _config.ConfiguredVirtualEnv(
                    name="a",
                    install=[
                        os.path.expanduser("~/a"),
                        os.path.expandvars("$HOME"),
                        os.path.expandvars("${HOME}/b"),
                    ],
                    requirements=[
                        os.path.expandvars("requirements-${HOME}.txt"),
                    ],
                ),
            ],
        )

    def test_explicit_python(self):
        config = _config.Config.from_string(
            """
            [virtualenv.a]
            install = ["foo"]
            python = "somepython2"
            """,
        )
        self.assertEqual(
            list(config), [
                _config.ConfiguredVirtualEnv(
                    name="a",
                    install=["foo"],
                    python="somepython2",
                ),
            ],
        )

    def test_duplicate_links(self):
        with self.assertRaises(_config.DuplicatedLinks) as e:
            _config.Config.from_string(
                """
                [virtualenv.a]
                link = ["python:foo", "pip:foo"]
                """,
            )
        self.assertIn("'foo'", str(e.exception))

    def test_duplicate_links_across_venvs(self):
        with self.assertRaises(_config.DuplicatedLinks) as e:
            _config.Config.from_string(
                """
                [virtualenv.a]
                link = ["python:foo"]
                [virtualenv.b]
                link = ["pip:foo"]
                """,
            )
        self.assertIn("'foo'", str(e.exception))

    def test_duplicate_link_modules(self):
        with self.assertRaises(_config.DuplicatedLinks) as e:
            _config.Config.from_string(
                """
                [virtualenv.a]
                link-module = ["pydoc:foo", "pip:foo"]
                """,
            )
        self.assertIn("'foo'", str(e.exception))

    def test_duplicate_link_modules_across_venvs(self):
        with self.assertRaises(_config.DuplicatedLinks) as e:
            _config.Config.from_string(
                """
                [virtualenv.a]
                link-module = ["pydoc:foo"]
                [virtualenv.b]
                link-module = ["pip:foo"]
                """,
            )
        self.assertIn("'foo'", str(e.exception))

    def test_duplicate_mixed_links(self):
        with self.assertRaises(_config.DuplicatedLinks) as e:
            _config.Config.from_string(
                """
                [virtualenv.a]
                link = ["python:foo"]
                link-module = ["pip:foo"]
                """,
            )
        self.assertIn("'foo'", str(e.exception))

    def test_duplicate_mixed_links_across_venvs(self):
        with self.assertRaises(_config.DuplicatedLinks) as e:
            _config.Config.from_string(
                """
                [virtualenv.a]
                link-module = ["pydoc:foo"]
                [virtualenv.b]
                link = ["pip:foo"]
                """,
            )
        self.assertIn("'foo'", str(e.exception))

    def test_add_with_contents(self):
        config = _config.Config()
        added = config.add("a", install=["foo", "bar"], link=["baz"])
        self.assertEqual(
            (list(config), list(added)),
            (
                [], [
                    _config.ConfiguredVirtualEnv(
                        name="a",
                        install=["foo", "bar"],
                        link={"baz": "baz"},
                    ),
                ],
            )
        )

    def test_add_empty(self):
        config = _config.Config()
        added = config.add("a")
        self.assertEqual(
            (list(config), list(added)),
            ([], [_config.ConfiguredVirtualEnv(name="a")]),
        )

    def test_empty(self):
        config = _config.Config()
        self.assertEqual(list(config), [])

    def test_empty_from_string(self):
        config = _config.Config.from_string("")
        self.assertEqual(config, _config.Config())
